use strict;
use vars qw($VERSION %IRSSI);

use Irssi;

$VERSION = '0.5';
%IRSSI = (
    authors     => 'cxreg',
    contact     => 'cxreg@pobox.com',
    name        => 'go',
    description => 'Switch to the window with the given name or item',
    license     => 'Public Domain',
    url         => 'http://genericorp.net/~count/irssi/go',
    changed     => '2008-01-29',
);

sub cmd_go {
    # Initialize the input bar
    _draw_suggestion('');

    # get a complete list of current windows
    my @all_windows = Irssi::windows();

    my ($window, $suggestion, @matches);
    my $buf = '';
    while(1) {
        # read input one character at a time
        my $chr = getc;

        # break out on Enter
        if ( $chr =~ /[\r\n]/ ) {
            $window = $suggestion;
            last;
        }

        if ( $chr =~ /[\t]/ ) {
            # get matches if we don't have any yet
            @matches = _match( $buf, @all_windows ) unless @matches;

            # if tab, cycle to next suggestion
            push @matches, grep { $_ } shift @matches;
            $suggestion = $matches[0];

            # draw suggestion
            _draw_suggestion( $buf, $suggestion );

            next;
        }

        # ^U means wipe out the input.  we might want to actually read this 
        # from the user's keybinding (for erase_line or maybe erase_to_beg_of_line)
        # instead of assuming ^U
        if ( ord($chr) == 21 ) {
            $buf = '';
            @matches = _match( $buf, @all_windows );
            $suggestion = undef;
        }

        # handle backspace and delete
        if ( ord($chr) == 127 or ord($chr) == 8 ) {
            # remove the last char
            $buf = substr( $buf, 0, length($buf) - 1 );

            # get suggestions again
            if ( @matches = _match( $buf, @all_windows ) ) {
                $suggestion = $buf ? $matches[0] : undef;
                _draw_suggestion( $buf, $suggestion );
            }
            else {
                # we should only get here if the entire input was deleted
                $suggestion = undef;
            }

            next;
        }

        # create a temporary new buffer
        my $tmp = $buf . $chr;

        if ( @matches = _match($tmp, @all_windows) ) {
            $buf = $tmp;
            $suggestion = $buf ? $matches[0] : undef;
        } else {
            # vbell on mistype
            print STDOUT "\a";
        }

        # display the current input and suggestion
        _draw_suggestion( $buf, $suggestion );
    }

    if ($window) {
        $window->{window}->set_active;
    }

    # refresh the screen to get the regular prompt back if needed
    Irssi::command('redraw')
}

Irssi::command_bind('go', 'cmd_go', 'go.pl');

# load the command as a keybinding?
# Irssi::command('/bind meta-w /go');

sub _draw_suggestion {
    my ( $b, $s ) = @_;

    my ($pre, $post, $extra) = (''x3);
    if ($s) {
        # No input, entire thing is a suggestion
        if ( !$b ) {
            $pre = '#' . $s->{window}->{refnum} . ' ';
            $post = $s->{string};
        }
        # Matched window number
        elsif ( $s->{match_type} eq 'number' and $s->{string} =~ /\Q$b\E/i ) {
            $pre = '#' . $`;
            $post = $' . ' ' . ( $s->{window}->{active}->{name} || $s->{window}->{name} );
        }
        # Matched 'tag' (network or server)
        elsif ( $s->{match_type} eq 'tag' and $s->{string} =~ /\Q$b\E/i ) {
            $pre = '#' . $s->{window}->{refnum} . ' ' . ( $s->{window}->{active}->{name} || $s->{window}->{name} ) . " ($`";
            $post = "$')";
        }
        # Matched window or item name
        elsif ( $s->{string} =~ /\Q$b\E/i ) {
            $pre = '#' . $s->{window}->{refnum} . ' ' . $`;
            $post = $';
        }

        # special case 'tag'.  maybe this should be moved up into the case blocks
        unless ( $s->{match_type} eq 'tag' ) {
            $extra = $s->{window}->{active_server}->{tag};
            $extra = " ($extra)" if $extra;
        }
    }

    # FIXME - there has to be a "right way" to do this.  
    # it looks like the fe-text/gui-readline.c and gui-entry.c
    # (and other gui-*) are not XS wrapped for whatever reason.
    print STDOUT "\r" . ' 'x40 . "\rGoto: ";

    print STDOUT "\x{1b}[7m$pre\x{1b}[0m" if $pre; # before 
    print STDOUT $b;                               # the matched string
    print STDOUT "\x{1b}[7m$post$extra\x{1b}[0m"   # after
        if $post or $extra;
}

sub _match {
    my ( $name, @wins ) = @_;
    my @matches;

    for (@wins) {
        if ( length $_->{name} and ( $_->{name} =~ /\Q$name\E/i or !length($name) ) ) {
            push @matches, { string => $_->{name}, window => $_, match_type => 'name' };
        }
        if ( length $_->{active}->{name} and 
                ( $_->{active}->{name} =~ /\Q$name\E/i or !length($name) ) ) {
            push @matches, { string => $_->{active}->{name}, window => $_, match_type => 'item' };
        }

        # don't add number or tag in absence of input, that's just too much redundancy
        if ( defined $_->{refnum} and $_->{refnum} =~ /\Q$name\E/i ) {
            push @matches, { string => $_->{refnum}, window => $_, match_type => 'number' };
        }
        if ( defined $_->{active_server}->{tag} and $_->{active_server}->{tag} =~ /\Q$name\E/i ) {
            push @matches, { string => $_->{active_server}->{tag}, window => $_, match_type => 'tag' };
        }
    }
    return @matches;
}
